# This code is part of the 'enjuewemela' game
# License: GPLv3
# Main author: Facundo Batista
# Code, bug tracker, etc:
#   https://launchpad.net/enjuewemela/
#
"""Gems for the game."""

from __future__ import division

import os

import cocos
from cocos.particle_systems import Galaxy
from cocos.particle import Color

import config


class _Gem(cocos.sprite.Sprite):
    """Gem wrapper around a sprite."""
    def __init__(self, filename, pos_id, explosive=False):
        cocos.sprite.Sprite.__init__(self, filename)
        # piece name is the file name without the .png
        self.name = filename[:-4]
        self.pos_id = pos_id
        self.is_explosive = explosive
        self.is_magic = False

    def get_explosive(self):
        """Return an explosive version of self."""
        if self.name[:4] == 'expl':
            new_name = '%s.png' % self.name
        else:
            new_name = 'expl-%s.png' % self.name
        g = _Gem(new_name, self.pos_id, explosive=True)
        g.scale = self.scale
        return g

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Gem %s at %d>" % (self.name, id(self))


class _Magic(Galaxy):
    """Magic ball."""
    def __init__(self):
        Galaxy.__init__(self)
        self.size = 20
        self.live = 10.0
        self.speed = 30
        self.radial_accel = -100
        self.tangential_accel = 30
        self.total_particles = 60
        self.start_color = Color(.7, .7, 1, 1)
        self.end_color = Color(1, 0, 0, .7)

        self.name = 'magic'
        self.pos_id = 13
        self.is_magic = True
        self.is_explosive = False

    def __str__(self):
        return self.name

    def __repr__(self):
        return "<Gem %s at %d>" % (self.name, id(self))

    def _set_scale(self, scale):
        self._scale = scale * 5
    def _get_scale(self):
        return self._scale / 5
    scale = property(_get_scale, _set_scale)


class Gems(object):
    """A gem representation."""

    def __init__(self):
        # load all the gems
        jewels = os.listdir(os.path.join(config.BASEDIR, "jewels"))
        self.pngs = sorted(x for x in jewels
                           if x.endswith('.png') and not x.startswith('expl'))
        if len(self.pngs) != 7:
            raise ValueError("The gems (png files) in the 'gems' directory "
                             "must be 7! (found %d)" % len(self.pngs))

        __, max_both = self._get_gems_size()

        # calculate the resize factor
        self.resize_factor = config.PIECE_SIZE / max_both

    def _get_gems_size(self):
        """Return the gems and the max size in both axis."""
        # convert pngs into sprites
        all_gems = [cocos.sprite.Sprite(x) for x in self.pngs]

        # get the maximum height or width of all of them
        max_w = max(x.width for x in all_gems)
        max_h = max(x.height for x in all_gems)
        max_both = max(max_w, max_h)
        return all_gems, max_both

    def get_all(self, size=None):
        """Get all the gems, resized if indicated."""
        all_gems, max_both = self._get_gems_size()

        if size is not None:
            # resize the sprites accordingly
            resize_factor = size / max_both
            for gem in all_gems:
                gem.scale = resize_factor
        return all_gems

    def get_magic_ball(self):
        """Return a magic ball gem."""
        sp = _Magic()
        sp.scale = self.resize_factor
        return sp

    def __getitem__(self, pos):
        sp = _Gem(self.pngs[pos], pos)
        sp.scale = self.resize_factor
        return sp

gems = Gems()
