# This code is part of the 'enjuewemela' game
# License: GPLv3
# Main author: Facundo Batista
# Code, bug tracker, etc:
#   https://launchpad.net/enjuewemela/
#
"""Module to keep and show game stats."""

from __future__ import division

import os

import cocos
from cocos.director import director
from cocos.text import Label

import config

import pyglet
from pyglet import font
font.add_directory(os.path.join(config.BASEDIR, 'fonts'))

import logic_board
from gems import gems
from sound import sounds
from levels import get_level_factors, FRUKIS


class Base(cocos.cocosnode.CocosNode):
    """White base for scores."""
    def __init__(self, x1, y1, width, height):
        super(Base, self).__init__()
        x2 = x1 + width
        y2 = y1 + height
        self.vertex = (x1, y1, x1, y2, x2, y2, x2, y1)

    def draw(self, *args, **kwargs):
        """Draw the white base."""
        pyglet.graphics.glPushMatrix()
        self.transform()
        pyglet.graphics.draw(4, pyglet.graphics.GL_QUADS,
                             ('v2i', self.vertex),
                             ('c4B', (255, 255, 255, 170)*4))
        pyglet.graphics.glPopMatrix()


class FontWriter(object):
    """Wrapper around font."""
    def __init__(self, layer, chosen_font):
        self.layer = layer
        self.font = chosen_font

    def __call__(self, text, position, size, anchor, color=(0, 0, 0, 255)):
        l = Label(text, position, font_name=self.font, font_size=size,
                  color=color, anchor_x=anchor, anchor_y="center")
        self.layer.add(l)
        return l


class Stats(cocos.layer.Layer):
    """Keeps score and statistics."""

    def __init__(self, game_manager):
        super(Stats, self).__init__()

        self.score = 0
        self.destroyed = dict.fromkeys(range(14), 0)
        self.added = dict.fromkeys(range(7), 0)
        self.bombs = 0
        self.magics = 0
        self.specials = 0
        self.level = 0
        self.level_factors = get_level_factors(0)
        self.frukis = 0  # FIXME: this depends of the game style
        self.gm = game_manager

        # the fonts to use
        f_titles = FontWriter(self, "Sansation")
        f_fixed = FontWriter(self, "SV Basic Manual")

        # some info for the initial drawing
        win_w, win_h = director.get_window_size()
        side_w = (win_w - config.ARISTA * 8) // 2
        base_w, base_h = 100, 480

        #
        # ------ left side ---------
        #
        # draw a white area centered in height and in the left border
        left_border = (side_w - base_w) // 2
        lower_y = (win_h - base_h) // 2
        b = Base(left_border, lower_y, base_w, base_h)
        self.add(b)

        middle = left_border + base_w // 2
        y = lower_y + base_h - 20

        # player
        f_titles("player", (middle, y), 9, "center",
                 color=(100, 127, 100, 255))
        y -= 20
        username = game_manager.config['username']
        f_titles(username, (middle, y), 15, "center", color=(0, 77, 105, 255))

        # level
        y -= 40
        f_titles(_("level"), (middle, y), 9, "center",
                 color=(100, 127, 100, 255))
        y -= 20
        self.level_label = f_fixed("0", (middle, y), 20, "center",
                                      color=(0, 0, 255, 255))

        # score
        y -= 40
        f_titles(_("score"), (middle, y), 9, "center",
                 color=(100, 127, 100, 255))
        y -= 20
        self.score_label = f_fixed("0", (middle, y), 20, "center",
                                      color=(127, 0, 127, 255))

        # get all pieces for level advance
        self.advance_pieces = gems.get_all()
        smiley = cocos.sprite.Sprite("smiley.png")
        self.advance_pieces.append(smiley)

        # precalculate the max scale for all pieces
        piece_max_size = 40
        max_len = max(max(p.height, p.width) for p in self.advance_pieces)
        self.final_advance_scale = piece_max_size / max_len

        # draw them according to needed size
        separation = (base_w - piece_max_size * 2) // 3
        y -= 50
        for i in range(4):
            y -= 50

            # left
            p = self.advance_pieces[i*2]
            p.position = (left_border + separation + piece_max_size // 2, y)
            p.scale = 0
            self.add(p)

            # right
            p = self.advance_pieces[i*2 + 1]
            p.position = (left_border + separation * 2 +
                          int(piece_max_size * 1.5), y)
            p.scale = 0
            self.add(p)

        # game time
        self.game_time_label = f_fixed("0:00", (middle, lower_y + 10), 14,
                                          "center", color=(250, 50, 50, 255))
        self.schedule_interval(self.update_timings, .5)


        #
        # ------ right side ---------
        #
        # draw a white area centered in height and in the right border
        left_border = win_w - (side_w - base_w) // 2 - base_w
        self.right_border = left_border + base_w
        lower_y = (win_h - base_h) // 2
        b = Base(left_border, lower_y, base_w, base_h)
        self.add(b)

        stats_heights = []

        # hits counter
        y = lower_y + base_h - 15
        f_titles(_("Hits:"), (left_border + 5, y), 12, "left")
        for t in _("Pieces: Bombs: Magics: Specials:").split():
            y -= 17
            stats_heights.append(y)
            f_titles(t, (left_border + 10, y), 10, "left")

        # added pieces
        y -= 35
        f_titles(_("Appeared:"), (left_border + 5, y), 12, "left")
        all_gems = gems.get_all(20)
        for i, gem in enumerate(all_gems):
            y -= 22
            stats_heights.append(y)
            gem.position = (left_border + 35, y)
            self.add(gem)

        # destroyed pieces
        y -= 35
        f_titles(_("Destroyed:"), (left_border + 5, y), 12, "left")
        all_gems = gems.get_all(20)
        for i, gem in enumerate(all_gems):
            y -= 22
            stats_heights.append(y)
            gem.position = (left_border + 35, y)
            self.add(gem)

        # initial values, keeping the labels to update later
        self.stats_texts = []
        for y in stats_heights:
            l = f_fixed("0", (self.right_border - 15, y), 12, "right")
            self.stats_texts.append(l)


    def redraw_stats(self):
        """Redraw the statistic values."""
        # these are the values to show
        to_show = [sum(self.destroyed.values()), self.bombs, self.magics,
                                                                self.specials]
        to_show.extend(self.added[x] for x in range(7))
        to_show.extend(self.destroyed[x] for x in range(7))

        # show them!
        for lab, val in zip(self.stats_texts, to_show):
            lab.element.text = str(val)

    def do_score(self, match_type, pieces, from_user):
        """Update stats for all match types."""
        print "======= do score", match_type, pieces, from_user
        positions = [p.position for p in pieces]
        if match_type == 'magic':
            self._do_score_magic(positions)
        else:
            if match_type == 'explosion':
                explosive = [p for p in pieces if p.is_explosive][0]
                center = explosive.position
            else:
                center = logic_board.get_center(positions)
            self._do_score_normal(positions, from_user, match_type, center)

    def _do_score_magic(self, positions):
        """Update stats when there is magic action."""
        base = len(positions) ** 2
        points = int(round(base * self.level_factors.lmp))
        each = points // len(positions)

        for position in positions:
            self.score_bubble(position, each)

        self.score += points
        self.score_label.element.text = str(int(self.score))
        self.frukis += points * self.level_factors.fpp
        self.redraw_stats()

    def _do_score_normal(self, positions, from_user, match_type, center):
        """Update stats when there is a gem match."""
        points = 0
        points += len(positions) * self.level_factors.lmp

        # adjust score accordingly to the length of match
        points *= len(positions)

        # if the match is because falling pieces, it'x 3x
        if not from_user:
            points *= 3

        # some special score
        if match_type == 'explode':
            points += 1.5

        # draw the bubble
        self.score_bubble(center, int(round(points)))

        # adjust score
        self.score += points
        self.score_label.element.text = str(int(self.score))
        self.frukis += points * self.level_factors.fpp
        self.redraw_stats()

    def score_bubble(self, position, points):
        """Shows a score that floats and fades away."""
        lab = Label(str(points), position, font_name="Komika Krak",
                    font_size=30, color=(0, 0, 55, 255), anchor_x="center")
        self.add(lab)

        def remove():
            """Removes the label."""
            self.remove(lab)

        lab.do(cocos.actions.FadeOut(config.BUBBLE_LIFE))
        lab.do(cocos.actions.MoveBy((0, 70), config.BUBBLE_LIFE))
        self.call_later(config.BUBBLE_LIFE, remove)

    def destroy_piece(self, piece):
        """Counts destroyed pieces."""
        self.destroyed[piece.pos_id] += 1

    def add_piece(self, piece):
        """Counts added pieces."""
        pos = piece.pos_id
        if pos < 13:
            self.added[pos] += 1

    def __str__(self):
        dt = self.gm.game_time
        return "Stats(time %d, score %d, added: %s, destroyed: %s)" % (
                                    dt, self.score, self.added, self.destroyed)

    def update_timings(self, dt):
        """Update game timings."""
        # time label
        time_text = "%d:%02d" % divmod(self.gm.game_time, 60)
        self.game_time_label.element.text = time_text

        # FIXME: make this work with timed games
        # adjust frukis regarding time
        # self.frukis -= ...
        if self.frukis < 0:
            self.game_over(timed=True)

        # need to check this here and not where we added frukis, because time
        # may have lowered it below the limit
        if self.frukis >= FRUKIS:
            sounds.play("new_level")
            self.frukis = 0  # FIXME: this depends of the game style
            self.level += 1
            self.level_factors = get_level_factors(self.level)
            self.level_label.element.text = str(self.level)

        # adjust the advance feedback
        fulls, graduated = divmod(self.frukis * 8, FRUKIS)
        for i, p in enumerate(self.advance_pieces):
            if i < fulls:
                p.scale = self.final_advance_scale
            elif i == fulls:
                p.scale =  self.final_advance_scale * graduated / FRUKIS
            else:
                p.scale = 0

    def game_over(self, timed=False, posib=False):
        """Game is over!"""
        sounds.play("game_over")
        if timed:
            print "=== GAME OVER POR TIEMPO!!!"
        elif posib:
            print "=== GAME OVER POR POSIBILIDADES!!!"
        else:
            raise ValueError("Don't know the reason of the game over!")

    def call_later(self, dt, func, *args, **kwargs):
        """Handy method to call later a function."""
        def f(dt, args, kwargs):
            """Unschedule the scheduled function."""
            self.unschedule(f)
            func(*args, **kwargs)
        self.schedule_interval(f, dt, args, kwargs)
